/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/


struct _DemoInfo {
	GtkFileSelector*	file_selector;
	int			mp3_id;
	int			mp3_mixer_id;
	int			top_mix;

	GtkWidget*		window;
	Gtk
	GList*			filters;
};

struct _FilterInfo {

	int			filter_id;
	DemoInfo*		demo_info;
	GtkWidget*		top_widget;

};


static DemoInfo*   demo_info_append_filter(DemoInfo*   demo_info,
                                           FilterInfo* filter_info)
{
	GList* last = g_list_last(demo_info->filters);
	GdamRequest** request = alloca(sizeof(GdamRequest*) * 2);
	GdamArg* arg = gdam_arg_new_from_valist(1,
	                                        "filtered",
				                GDAM_ARG_TYPE_INT,
				                last ? last->filter_id
				                     : demo_info->mp3_mixer_id);
	request[0] = gdam_request_set_args(filter_info->filter_id,
	                                   1,
					   arg,
					   (GHookFunc) do_nothing,
					   (GHookFunc) do_nothing,
					   NULL);
	request[1] = gdam_request_add_mix(demo_info->mixer_id,
			                  filter_info->filter_id,
					  -1,
					  (GHookFunc) do_nothing,
					  (GHookFunc) do_nothing,
					  NULL);
	package_request = gdam_request_package_v(2, request, 
					 (GdamRequestGotIntFunc) do_nothing,
					         (GHookFunc) request_failed,
						 NULL);
	if (!gdam_channel_process_request(channel, package_request))
		g_error("error talking to server");
}

static DemoInfo*   demo_info_prepend_filter(DemoInfo*   demo_info,
                                            FilterInfo* filter_info)
{
	GdamRequest** request = alloca(sizeof(GdamRequest*) * 2);
	GdamArg* arg = gdam_arg_new_from_valist(1,
	                                        "filtered",
				                GDAM_ARG_TYPE_INT,
				                last->filter_id);
	request[0] = gdam_request_set_args(filter_info->filter_id,
	                                   1,
					   arg,
					   (GHookFunc) do_nothing,
					   (GHookFunc) do_nothing,
					   NULL);
	if (demo_info->filters) {
		FilterInfo* old_first = (FilterInfo*)demo_info->filter->data;
		GdamArg* farg
		farg = gdam_arg_new_from_valist(1,
		                                "filtered",
				                GDAM_ARG_TYPE_INT,
						filter_info->filter_id);
		request[1] = gdam_request_set_args(old_first->filter_id,
		                                   1,
						   farg,
					           (GHookFunc) do_nothing,
					           (GHookFunc) do_nothing,
					           NULL);
	} else {
		request[1] = gdam_request_add_mix(demo_info->mixer_id,
						  filter_info->filter_id,
						  -1,
						  (GHookFunc) do_nothing,
						  (GHookFunc) do_nothing,
						  NULL);
	}
	package_request = gdam_request_package_v(2, request, 
					 (GdamRequestGotIntFunc) do_nothing,
					         (GHookFunc) request_failed,
						 NULL);
	if (!gdam_channel_process_request(channel, package_request))
		g_error("error talking to server");
}

static DemoInfo*        demo_info_new(GdamChannel* channel, int mixer_id)
{
	DemoInfo* info = g_new0(DemoInfo, 1);
	char* song;
	info->file_selection = GTK_FILE_SELECTION(
				gtk_file_selection_new("Base mp3")
	);
	info->mp3_id = 0;
	info->mp3_mixer_id = gdam_channel_create_source_generic(
					channel,
					"GdamMix",
					1,
					"terminates",
					GDAM_ARG_TYPE_INT,
					0);
	info->top_mix = mixer_id;
	info->window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	info->filters = NULL;

	song = g_getenv("GDAMDEMO");
	if (song == NULL)
		song = "???";
	info->label = gtk_label_new(song);

	return info;
}
