#include <gdam/component/gdamsongdatabase.h>
#include <gdam/gdamtildeexpand.h>
#include "gdamsongdatabasemakerskin.h"
#include "gdammp3info.h"
#include "../gdamskinutil.h"
#include <gtk/gtksignal.h>
#include <sys/wait.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>

static void handle_find_is_name_toggle(GtkToggleButton* button, 
                          GdamSongDatabaseMakerSkin* skin)
{
	int is_active = gtk_toggle_button_get_active(button) ? 1 : 0;
	skin->find_is_name = is_active;
}
static gboolean gdam_song_database_maker_find_is_name_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	GtkToggleButton* button = GTK_TOGGLE_BUTTON(object);
	int is_active;
	g_return_val_if_fail(GTK_IS_TOGGLE_BUTTON(object), TRUE);
	is_active = gtk_toggle_button_get_active(button) ? 1 : 0;
	GDAM_SONG_DATABASE_MAKER_SKIN(skin)->find_is_name = is_active;
	gtk_signal_connect(object,
	                   signal_name,
	                   GTK_SIGNAL_FUNC(handle_find_is_name_toggle),
			   skin);
	return TRUE;
}

static gboolean gdam_song_database_maker_expression_entry_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	GtkEntry* entry = GTK_ENTRY(object);
	g_return_val_if_fail(entry != NULL, TRUE);
	GDAM_SONG_DATABASE_MAKER_SKIN(skin)->find_name_or_regex = entry;
	return TRUE;
}
static gboolean gdam_song_database_maker_find_path_entry_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	GtkEntry* entry = GTK_ENTRY(object);
	g_return_val_if_fail(entry != NULL, TRUE);
	GDAM_SONG_DATABASE_MAKER_SKIN(skin)->find_starting_path = entry;
	return TRUE;
}

static void handle_traverse_mount_point_toggle(GtkToggleButton* button, 
                          GdamSongDatabaseMakerSkin* skin)
{
	int is_active = gtk_toggle_button_get_active(button) ? 1 : 0;
	skin->find_traverse_mount_points = is_active;
}
static gboolean gdam_song_database_maker_traverse_mount_point_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	GtkToggleButton* button = GTK_TOGGLE_BUTTON(object);
	int is_active;
	g_return_val_if_fail(GTK_IS_TOGGLE_BUTTON(object), TRUE);
	is_active = gtk_toggle_button_get_active(button) ? 1 : 0;
	GDAM_SONG_DATABASE_MAKER_SKIN(skin)->find_traverse_mount_points
					= is_active;
	gtk_signal_connect(object,
	                   signal_name,
	                   GTK_SIGNAL_FUNC(handle_traverse_mount_point_toggle),
			   skin);
	return TRUE;
}
static void handle_traverse_symlink_toggle(GtkToggleButton* button, 
                          GdamSongDatabaseMakerSkin* skin)
{
	int is_active = gtk_toggle_button_get_active(button) ? 1 : 0;
	skin->find_traverse_symlinks = is_active;
}
static gboolean gdam_song_database_maker_traverse_symlink_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	GtkToggleButton* button = GTK_TOGGLE_BUTTON(object);
	int is_active;
	g_return_val_if_fail(GTK_IS_TOGGLE_BUTTON(object), TRUE);
	is_active = gtk_toggle_button_get_active(button) ? 1 : 0;
	GDAM_SONG_DATABASE_MAKER_SKIN(skin)->find_traverse_symlinks = is_active;
	g_return_val_if_fail(GTK_IS_TOGGLE_BUTTON(object), TRUE);
	gtk_signal_connect(object,
	                   signal_name,
	                   GTK_SIGNAL_FUNC(handle_traverse_symlink_toggle),
			   skin);
	return TRUE;
}

static void do_run_find(GdamSongDatabaseMakerSkin* skin)
{
	gdam_song_database_maker_run_find(skin);
}

static gboolean gdam_song_database_maker_run_find_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	gtk_signal_connect_object(object,
	                          signal_name,
				  GTK_SIGNAL_FUNC(do_run_find),
				  GTK_OBJECT(skin));
	return TRUE;
}

static gboolean gdam_song_database_maker_locate_entry_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	GtkEntry* entry = GTK_ENTRY(object);
	g_return_val_if_fail(entry != NULL, TRUE);
	GDAM_SONG_DATABASE_MAKER_SKIN(skin)->locate_entry = entry;
	return TRUE;
}

static void do_run_locate(GdamSongDatabaseMakerSkin* skin)
{
	gdam_song_database_maker_run_locate(skin);
}

static gboolean gdam_song_database_maker_run_locate_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	gtk_signal_connect_object(object,
	                          signal_name,
				  GTK_SIGNAL_FUNC(do_run_locate),
				  GTK_OBJECT(skin));
	return TRUE;
}

static gboolean gdam_song_database_maker_arbitrary_entry_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	GtkEntry* entry = GTK_ENTRY(object);
	g_return_val_if_fail(entry != NULL, TRUE);
	GDAM_SONG_DATABASE_MAKER_SKIN(skin)->arbitrary_entry = entry;
	return TRUE;
}

static void do_run_arbitrary(GdamSongDatabaseMakerSkin* skin)
{
	gdam_song_database_maker_run_arbitrary(skin);
}

static gboolean gdam_song_database_maker_run_arbitrary_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	gtk_signal_connect_object(object,
	                          signal_name,
				  GTK_SIGNAL_FUNC(do_run_arbitrary),
				  GTK_OBJECT(skin));
	return TRUE;
}


static GdamSkinGenericConnect connect_infos[] =
{	{ "gdam_song_database_maker_find_is_name",
          gdam_song_database_maker_find_is_name_handler		},
	{ "gdam_song_database_maker_expression_entry",
          gdam_song_database_maker_expression_entry_handler	},
	{ "gdam_song_database_maker_find_path_entry",
          gdam_song_database_maker_find_path_entry_handler	},
	{ "gdam_song_database_maker_traverse_mount_point",
          gdam_song_database_maker_traverse_mount_point_handler	},
	{ "gdam_song_database_maker_traverse_symlink",
          gdam_song_database_maker_traverse_symlink_handler	},
	{ "gdam_song_database_maker_run_find",
          gdam_song_database_maker_run_find_handler		},
	{ "gdam_song_database_maker_locate_entry",
          gdam_song_database_maker_locate_entry_handler		},
	{ "gdam_song_database_maker_run_locate",
          gdam_song_database_maker_run_locate_handler		},
	{ "gdam_song_database_maker_arbitrary_entry",
          gdam_song_database_maker_arbitrary_entry_handler	},
	{ "gdam_song_database_maker_run_arbitrary",
          gdam_song_database_maker_run_arbitrary_handler	},	};

static void gdam_song_database_maker_skin_init(GdamSongDatabaseMakerSkin* skin)
{
	skin->pid_table = g_hash_table_new(NULL, NULL);
}
static void gdam_song_database_maker_skin_class_init(GdamSkinClass* skin_class)
{
	gdam_skin_class_register_generic_handlers(skin_class,
				                  G_N_ELEMENTS(connect_infos),
				                  connect_infos);
}

GtkType gdam_song_database_maker_skin_get_type()
{
	static GtkType song_database_maker_skin_type = 0;
	if (!song_database_maker_skin_type) {
		static const GtkTypeInfo song_database_maker_skin_info = {
			"GdamSongDatabaseMakerSkin",
			sizeof(GdamSongDatabaseMakerSkin),
			sizeof(GdamSongDatabaseMakerSkinClass),
			(GtkClassInitFunc) gdam_song_database_maker_skin_class_init,
			(GtkObjectInitFunc) gdam_song_database_maker_skin_init,
			/* reserved_1*/ NULL,
			/* reserved_2*/ NULL,
			(GtkClassInitFunc) NULL
		};
		GtkType parent = GDAM_TYPE_SKIN;
		song_database_maker_skin_type = gtk_type_unique(parent,
		                                 &song_database_maker_skin_info);
	}
	return song_database_maker_skin_type;
}

static void build_song_entry(GdamSongDatabase* database, const char* filename)
{
	GdamSongInfo* song_info;
	GdamMp3Info* mp3_info;
	song_info = gdam_song_database_find(database, filename);
	if (song_info != NULL) {
		/* Don't requery the file... */
		return;
	}
	mp3_info = gdam_mp3_info_new(filename);
	if (mp3_info == NULL) {
		g_warning("%s was not an mp3 file", filename);
		return;
	}
	//g_message("building song entry for %s", filename);
	song_info = gdam_song_info_new(filename, 0.0, 0.0);
	gdam_song_info_set_length(song_info, mp3_info->length);
        /*  gboolean gdam_song_info_add_property(GdamSongInfo*     song_info,
                                                 const char*       key,
						 const char*       value); */
	gdam_song_database_add(database, song_info);
	gdam_mp3_info_destroy(mp3_info);
}

static gboolean process_one_song(gpointer user_data)
{
	GdamSongDatabaseMakerSkin* maker = user_data;
	char* fname;
	GdamSongDatabase* database;
	if (maker->first_pending_file == NULL) {
		maker->pending_thread = 0;
		return FALSE;
	}
	fname = maker->first_pending_file->data;
	//g_message("process_one_song: %s", fname);
	{
		GSList* to_free = maker->first_pending_file;
		maker->first_pending_file = to_free->next;
		if (to_free->next == NULL)
			 maker->last_pending_file = NULL;
	 	g_slist_free_1(to_free);
	}
	database = gdam_song_database_system();
	g_return_val_if_fail(database != NULL, FALSE);
	if (gdam_song_database_find(database, fname)) {
		g_free(fname);
	} else {
		build_song_entry(database, fname);
	}
	if (maker->first_pending_file == NULL) {
		maker->pending_thread = 0;
		gtk_signal_emit_by_name(GTK_OBJECT(database), "changed");
	}
	return maker->first_pending_file != NULL;
}

void  gdam_song_database_maker_add_song
                              (GdamSongDatabaseMakerSkin* skin,
			       const char*                filename)
{
	GSList* list = g_slist_prepend(NULL, g_strdup(filename));
	if (skin->first_pending_file == NULL)
		skin->first_pending_file = list;
	else
		skin->last_pending_file->next = list;
	skin->last_pending_file = list;
	if (skin->pending_thread == 0)
		skin->pending_thread = g_idle_add(process_one_song, skin);
}

typedef struct _FileListThreadData FileListThreadData;
struct _FileListThreadData {
	int			pid;
	GIOChannel*		io_channel;
	GdamSongDatabaseMakerSkin* skin;
	GdamBuffer		line_buffer;
	char			separator;
};

static void add_song_with_name_from_buffer
                              (GdamSongDatabaseMakerSkin* skin,
			       GdamBuffer*                buffer,
			       int                        index)
{
	char* str = alloca(index + 1);
	gdam_buffer_read(buffer, str, index +1);
	str[index] = '\0';
	g_message("add song %s", str);
	gdam_song_database_maker_add_song(skin, str);
}

static void cleanup_thread_data(FileListThreadData* thread_data)
{
	int status;
	GdamSongDatabaseMakerSkin* skin = thread_data->skin;
	gdam_buffer_destruct(&thread_data->line_buffer);
	kill(thread_data->pid, SIGKILL);
	g_hash_table_remove(skin->pid_table,
	                    GINT_TO_POINTER(thread_data->pid));
	if (waitpid(thread_data->pid, &status, 0) < 0) {
		g_warning("waitpid: pid %d: %s",
			thread_data->pid,
			strerror(errno));
	}
	gtk_object_unref(GTK_OBJECT(skin));
	g_free(thread_data);
}
static int do_fork(char** command, int* fd_out)
{
	int fds[2];
	int pid;
	if (pipe(fds) < 0) {	
		g_warning("do_fork: couldn't create pipe: %s",
			strerror(errno));
		return 0;
	}
	while ((pid = fork()) < 0) {
		if (errno == EAGAIN) { sleep(5); continue; }
		if (errno == EINTR) { continue; }
		g_warning("fork failed: %s", strerror(errno));
		return 0;
	}
	if (pid == 0) {
		close(fds[0]);
		dup2(fds[1], STDOUT_FILENO);
		execvp(command[0], command);
		_exit(127);
	}
	close(fds[1]);
	*fd_out = fds[0];
	return pid;
}

static gboolean process_filename_list(GIOChannel   *source,
				      GIOCondition  condition,
			              gpointer      data)
{
	GIOError io_error;
	char buf[4096];
	int num_read;
	int index;
	FileListThreadData* thread_data = data;
	GdamBuffer* buffer;
	GdamSongDatabaseMakerSkin* skin = thread_data->skin;
	io_error = g_io_channel_read(source, buf, sizeof(buf), &num_read);
	if (io_error != G_IO_ERROR_NONE) {
		g_warning("ioerror");
		cleanup_thread_data(thread_data);
		return FALSE;
	}
	//g_message("got %d bytes for external song database making command", num_read);
	buffer = &thread_data->line_buffer;
	gdam_buffer_append(buffer, buf, num_read);
	while ((index=gdam_buffer_index_of(buffer,
	                                   thread_data->separator,
					   0)                    ) >= 0) {
		add_song_with_name_from_buffer(skin, buffer, index);
	}
	if (num_read <= 0) {
		cleanup_thread_data(thread_data);
		return FALSE;
	}
	//g_io_channel_unref(thread_data->io_channel);
	return TRUE;
}

guint gdam_song_database_maker_run_command
                              (GdamSongDatabaseMakerSkin* skin,
                               char**                     command,
			       gboolean                   make_update_window,
			       char                       separator)
{
	FileListThreadData* thread_data;
	int pid, fd;
	thread_data = g_new(FileListThreadData, 1);
	pid = do_fork(command, &fd);
	g_message("ran command: %s %s", command[0], command[1]);
	if (pid == 0) {
		g_warning("couldn't run command");
		return 0;
	}
	g_return_val_if_fail(fd >= 0, 0);
	gtk_object_ref(GTK_OBJECT(skin));
	thread_data->skin = skin;
	thread_data->io_channel = g_io_channel_unix_new(fd);
	thread_data->pid = pid;
	thread_data->separator = separator;
	gdam_buffer_construct(&thread_data->line_buffer);
	g_io_add_watch(thread_data->io_channel,
	               G_IO_IN | G_IO_HUP,
		       process_filename_list,
		       thread_data);
	g_hash_table_insert(skin->pid_table, GINT_TO_POINTER(pid), thread_data);
	return pid;
}

guint gdam_song_database_maker_run_find(GdamSongDatabaseMakerSkin* skin)
{
	char* start_path = NULL;
	char* name = "*.mp3";
	char* command[20];
	char* expanded_path = NULL;
	char* path;
	int i;

	if (skin->find_starting_path != NULL) {
		start_path = gtk_entry_get_text(skin->find_starting_path);
		expanded_path = gdam_tilde_expand(start_path);
	}
	if (start_path == NULL) {
		start_path = "/";
	}
	if (expanded_path)
		path = expanded_path;
	else
		path = start_path;

	if (skin->find_name_or_regex != NULL)
		name = gtk_entry_get_text(skin->find_name_or_regex);

	command[0] = "find";
	command[1] = path;
	i = 2;
	if (skin->find_traverse_symlinks)
		command[i++] = "-follow";
	if (! skin->find_traverse_mount_points)
		command[i++] = "-xdev";
	command[i++] = (skin->find_is_name ? "-name" : "-regex");
	command[i++] = name;
	command[i] = NULL;
	return gdam_song_database_maker_run_command(skin, command, TRUE, '\n');
}

guint gdam_song_database_maker_run_locate(GdamSongDatabaseMakerSkin* skin)
{
	char* command[3];
	char* name = ".mp3";
	if (skin->locate_entry != NULL)
		name = gtk_entry_get_text(skin->locate_entry);
	command[0] = "locate";
	command[1] = name;
	command[2] = NULL;
	return gdam_song_database_maker_run_command(skin, command, TRUE, '\n');
}

guint gdam_song_database_maker_run_arbitrary(GdamSongDatabaseMakerSkin* skin)
{
	char* command[4];
	char* name;
	if (skin->locate_entry == NULL) {
		g_warning("run arbitrary command: no entry box");
		return 0;
	}
	name = gtk_entry_get_text(skin->arbitrary_entry);
	g_message("arbi: %s", name);
	command[0] = "/bin/sh";
	command[1] = "-c";
	command[2] = name;
	command[3] = NULL;
	return gdam_song_database_maker_run_command(skin, command, TRUE, '\n');
}

void  gdam_song_database_maker_kill(GdamSongDatabaseMakerSkin* skin, guint id)
{
	FileListThreadData* thread_data;
	thread_data = g_hash_table_lookup(skin->pid_table,
	                                  GUINT_TO_POINTER(id));
	if (thread_data != NULL)
		cleanup_thread_data(thread_data);
}
